from enum import StrEnum
from typing import Any


class CaseInsensitiveEnum(StrEnum):
    @classmethod
    def _missing_(cls, value):
        value_str = str(value)  # Ensure the value is a string
        for member in cls:
            if member.value.lower() == value_str.lower():
                return member
        raise ValueError(f"{value} is not a valid {cls.__name__}")


class APIType(CaseInsensitiveEnum):
    OTHER = "Other"
    ANTHROPIC = "Anthropic"
    OPENAI = "OpenAI"

    @staticmethod
    def from_model(model: str) -> "APIType":
        model = model.lower()

        if model in AnthropicModel.__members__.values():
            return APIType.ANTHROPIC
        if model in OpenAIModel.__members__.values():
            return APIType.OPENAI
        return APIType.OTHER


class APIModel(CaseInsensitiveEnum):
    def __str__(self):
        return self.value

    def __repr__(self) -> str:
        return self.value

    @classmethod
    def from_string(cls, value: str) -> "APIModel":
        return cls(value)

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type: Any, _handler: Any):
        from pydantic_core import core_schema

        def validate(value):
            if isinstance(value, cls):
                return value
            if isinstance(value, str):
                return cls(value)
            raise ValueError(f"Invalid value for {cls.__name__}: {value}")

        return core_schema.no_info_plain_validator_function(validate)


class AnthropicModel(APIModel):
    CLAUDE_3_SONNET_20240229 = "claude-3-sonnet-20240229"
    CLAUDE_3_5_SONNET_20240620 = "claude-3-5-sonnet-20240620"
    CLAUDE_3_5_SONNET_20241022 = "claude-3-5-sonnet-20241022"
    CLAUDE_3_HAIKU_20240307 = "claude-3-haiku-20240307"
    CLAUDE_3_OPUS_20240229 = "claude-3-opus-20240229"


class OpenAIModel(APIModel):
    DAVINCI = "davinci"
    CURIE = "curie"
    GPT_3_5_TURBO_0125 = "gpt-3.5-turbo-0125"
    GPT_3_5_TURBO = "gpt-3.5-turbo"
    GPT_3_5_TURBO_INSTRUCT = "gpt-3.5-turbo-instruct"
    GPT_4 = "gpt-4"
    GPT_4O = "gpt-4o"
    GPT_4O_MINI = "gpt-4o-mini"
    GPT_4O_2024_05_13 = "gpt-4o-2024-05-13"
    O1_MINI = "o1-mini"
    O1_PREVIEW = "o1-preview"



from dataclasses import dataclass
from typing import Any, Callable, Optional

import instructor

from qrev_instructor.models import AnthropicModel, APIModel, APIType, OpenAIModel


## Load the APIs
@dataclass
class FuncApiWrapper:
    """
    Wrapper for the API function call and class
    """

    from_class: type[Any]
    from_func: Callable[..., Any]
    original_func: Optional[Callable[..., Any]] = None


def get_apis() -> dict[APIType, FuncApiWrapper]:
    # Load the OpenAI API
    apis: dict[APIType, FuncApiWrapper] = {}
    try:
        from openai import OpenAI

        apis[APIType.OPENAI] = FuncApiWrapper(OpenAI, instructor.from_openai)

    except ImportError:
        pass

    # Load the Anthropic API
    try:
        from anthropic import Anthropic

        apis[APIType.ANTHROPIC] = FuncApiWrapper(Anthropic, instructor.from_anthropic)

    except ImportError:
        pass
    return apis



def get_api_service(model_name: str) -> APIType:
    model_name = model_name.lower()
    if model_name in AnthropicModel._value2member_map_:
        return APIType.ANTHROPIC
    elif model_name in OpenAIModel._value2member_map_:
        return APIType.OPENAI
    model_name = model_name.replace("-", "_").replace(".", "_").upper()
    if model_name in AnthropicModel._member_names_:
        return APIType.ANTHROPIC
    elif model_name in OpenAIModel._member_names_:
        return APIType.OPENAI
    else:
        raise NotImplementedError(f"Model {model_name} not supported")



import instructor as instructor

from qrev_instructor.client import get_api_service, get_client
from qrev_instructor.maybe import Maybe, MaybeBase
from qrev_instructor.models import AnthropicModel, APIModel, APIType, OpenAIModel

__all__ = [
    "get_api_service",
    "get_client",
    "AnthropicModel",
    "APIModel",
    "APIType",
    "OpenAIModel",
    "instructor",
    "Maybe",
    "MaybeBase",
]



from typing import Dict, Literal, TypedDict, Union

try:
    from openai.types.chat import ChatCompletionMessageParam  # type: ignore
except ImportError:
    ChatCompletionMessageParam = Dict[str,str] # type: ignore

def to_user_message(s: str) -> ChatCompletionMessageParam: # type: ignore
    return {
        "role": "user",
        "content": s,
    }

def to_system_message(s: str) -> ChatCompletionMessageParam: # type: ignore
    return {
        "role": "system",
        "content": s,
    }

def to_assistant_message(s: str) -> ChatCompletionMessageParam: # type: ignore
    return {
        "role": "assistant",
        "content": s,
    }

def to_tool_message(s: str) -> ChatCompletionMessageParam: # type: ignore
    return { # type: ignore
        "role": "tool",
        "content": s,
    }

def to_function_message(s: str) -> ChatCompletionMessageParam: # type: ignore
    return { # type: ignore
        "role": "function",
        "content": s,
    }


from typing import Generic, Optional

from instructor.dsl.maybe import T
from pydantic import BaseModel, Field, create_model


class MaybeBase(BaseModel, Generic[T]):
    """
    Extract a result from a model, if any, otherwise set the error and message fields.
    Mostly copies from instructor.maybe.MaybeBase.
    These fields do get overridden by the fields in the Maybe function, but specified here for typing.
    """

    result: T
    extraction_clarifying_message: Optional[str] = Field(
        default=None, description="Ask the user for clarification if for any fields that "
    )

    def __bool__(self) -> bool:
        return self.result is not None

    @property
    def error(self) -> bool:
        return self.result is None


def Maybe(
    model: type[T], result_description: Optional[str] = None, ecm_description: Optional[str] = None
) -> type[MaybeBase[T]]:
    ## Get required fields
    fields = list(model.model_fields.keys())
    required_fields = [field for field in fields if model.model_fields[field].is_required()]
    ## Descriptions
    result_description = (
        result_description or "Correctly extracted result from the model, if any, otherwise None"
    )
    ecm_description = ecm_description or (
        f"A message asking the user to provide more information when any of "
        f"these fields are missing or unclear, [{','.join(required_fields)}]. "
        f"This message should guide the user to specify the necessary "
        f"details."
    )
    ## capitalize the first letter of the model name
    name = model.__name__[0].upper() + model.__name__[1:]

    ## Field names here (result, extraction_clarifying_message) override the ones in MaybeBase
    return create_model(
        f"Maybe{name}",
        __base__=MaybeBase, # type: ignore
        result=(Optional[model], Field(default=None, description=result_description)),
        extraction_clarifying_message=(str, Field(default=None, description=ecm_description)),
    )



from typing import Any, Iterable, Optional, TypeVar, Union

import instructor
from pydantic import BaseModel

from qrev_instructor.apis import get_api_service, get_apis
from qrev_instructor.models import APIModel, APIType

T = TypeVar("T", bound=Union[BaseModel, "Iterable[Any]"])

apis = get_apis()


def _get_client(
    model: Optional[str | APIModel] = None,
    client: Optional[Any] = None,
    llm_type: Optional[str] = None,
    client_params: Optional[dict[str, Any]] = None,
    func_params: Optional[dict[str, Any]] = None,
) -> instructor.Instructor:
    if model and llm_type:
        raise ValueError(f"Cannot specify both model and llm_type. Got {model} and {llm_type}")
    if client:
        client_cls_name = client.__class__.__name__.lower()
        if "anthropic" in client_cls_name:
            _llm_type = APIType.ANTHROPIC
        elif "openai" in client_cls_name:
            _llm_type = APIType.OPENAI
        else:
            raise NotImplementedError(f"Client {client} not supported yet")
    else:
        _llm_type = None
    nllm_type = None
    if model:
        nllm_type = get_api_service(str(model))
    elif llm_type:
        nllm_type = APIType(llm_type)
    if _llm_type and nllm_type and nllm_type != _llm_type:
        raise ValueError(f"Cannot specify different llm_types. Got {_llm_type} != {nllm_type}")
    if _llm_type is None:
        _llm_type = nllm_type
    if nllm_type is None:
        nllm_type = APIType.OPENAI

    if client_params is None:
        client_params = {}
    if func_params is None:
        func_params = {}
    wrap = apis[nllm_type]
    instance = wrap.from_class(**client_params)
    if client:
        return_client = wrap.from_func(client, **func_params)
    else:
        return_client = wrap.from_func(instance, **func_params)
    return return_client


def get_client(
    model: Optional[str | APIModel] = None,
    client: Optional[Any] = None,
    llm_type: Optional[str] = None,
    client_params: Optional[dict[str, Any]] = None,
    func_params: Optional[dict[str, Any]] = None,
) -> instructor.Instructor:
    client = _get_client(model=model, client=client, llm_type=llm_type, client_params=client_params,func_params= func_params)
    return client
